#ifndef __MESH__
#define __MESH__

#include "common.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Class Mesh
//
// This class is holds the current delaunay triangulation of the mesh.
//
//////////////////////////////////////////////////////////////////////

class Mesh {
public:
   int dim;             // dimension of the mesh...
   int id;          // id from the data file
   int gid;             // unique numerical id
   string name;         // for labeling

   double bbox;         // size of the bounding box...
   double upperbound[3];
   double lowerbound[3];
   bool convex;
   bool protect;

   PLC *C;

   map < int, Simplex * >p2s;   // map from point to some simplex containing it...

   int np;          // number of points
   int ns;          // number of simplices...

   Vector dir;          // direction of the line or normal to the plane
   Point base;          // a reference point in this mesh
   Collar *col;
   Intestine * intestine;
   Boundary *bnd;

   map < int, int >ip2nb;
   set < Mesh * >bndseg;

// pointers to associated triangulations
   vector < Mesh * >children;
   vector < Mesh * >parents;

   set < int >exteriorPoints;

// constructor...
   Mesh(PLC * complex);

// write destructor...

   bool IsIn(Mesh * m);
   bool Contains(Mesh * m);
// high level operations...
   void Neighbors(int i, set < int >&nbs);
   void SegNeighbors(int i, int & n1, int & n2);
   void Simplices(int p, set < Simplex * >&star);

   bool Exists(int i);
   void Exists(Simplex * s);
   Simplex *GetSimplex();
   Simplex *GetSimplex(int i);
   int GetPoint();

   bool Exterior(Simplex * s);
   bool Exterior(int p);

   bool IsBoundary(int p);
   bool IsBndHelp(int bp, int pa, int pb);

   bool IsCollar(Simplex * s);
   bool IsCollarHelp(int cp, int pa, int pb);
   bool IsIntestine(Simplex * s);

   bool IsIntestineEdge(int a, int b);

   void GetSegmentEndPoints(int &a, int &b);
   bool IsEndSegment(Simplex * s);

// find the triangle that contains p.
// currently just finds a triangle that p encroaches upon...
   Simplex *Locate(Point p);
   Simplex *Locate(int i);
// hint is a starting point...
   Simplex *Locate(Point p, int hint);
   Simplex *Locate(int i, int hint);
// s is a starting simplex...
   Simplex *Locate(Point p, Simplex * s);
   Simplex *Locate(int i, Simplex * s);

// clear the cavity and add new triangles
// s is the starting simplex for the search for i
// cavity returns all new simplices added...
   void InsertPoint(int i);
   void InsertPoint(int i, int hint);
   void InsertPoint(int i, Simplex * s);
   void InsertPoint(int i, set < Simplex * >&oldcavity,
                    set < Simplex * >&newcavity);
   void InsertPoint(int i, int hint, set < Simplex * >&oldcavity,
                    set < Simplex * >&newcavity);
   void InsertPoint(int i, Simplex * s, set < Simplex * >&oldcavity,
                    set < Simplex * >&newcavity);

// Utility Function...
// Insert Point into triangulation but do not update p2s...
// make this private? (2-18-08)
   void InsertPointNoP2S(int i, Simplex * s, set < Simplex * >&oldcavity,
                         set < Simplex * >&newcavity);

// FIXME: this has not been implemented...
   void RemovePoint(int i, set < Simplex * >&cavity);
   void RemovePoint(int i, set < Simplex * >&oldcavity,
                    set < Simplex * >&cavity);

// return the cavity and the boundary...
   void Cavity(int p, Simplex * t, set < Simplex * >&cavity,
               set < Face > &bndry);
   void Cavity(Point & p, Simplex * t, set < Simplex * >&cavity,
               set < Face > &bndry);
   void Cavity(Point & p, set < int >&nbs);
   void Cavity(Point & p, Simplex * t, set < int >&nbs);

// add the new tets with the new point
   void Stitch(int i, set < Face > &bndry, set < Simplex * >&newcavity);

   int NP();
   int NS();

   int SegmentFirstPoint();     // get a bbox point for a segment...
   int SegmentNextPoint(int cur, int prev);     // walk down a segment...

   bool IsChild(Mesh * m1);

   void NormalInFace(int a, int b, Vector & n);
   void NormalInFace(Mesh * mSeg, Vector & n);

   void SetRadiusAndCC(Simplex * s);


// print .ply file...
   void PrintOff(string fname);
   void PrintCollarOff(string fname);
   void PrintSurfaceOff(string fname);
   void PrintSurfaceGMV(string fname);
   void PrintSurfaceCOff(string fname);
   void PrintEPS(string fn);  // for 2D stuff...
   void PrintSurfaceEPS(string fn);  // for 2D stuff...

   void PrintInfo();

// set up the bounding box...
   void Initialize(vector < Simplex * >&initialSimplices);

   void DeleteMesh();

   void Setp2s(Simplex * s);

   class Iterator {
public:
      Iterator();
      void reset();
      void operator++();
      void Start(Simplex * s);  // JUNK: 9-19-08.  remove this soon
      Simplex *Current();

private:
      set < Simplex * >visited;
      queue < Simplex * >Q;
   };

   class PIterator {
public:
      PIterator(Mesh * mm);
      void reset();
      void operator++();
      void Start(Simplex * s);  // JUNK: 9-19-08.  remove this soon
      int Current();

private:
//set<int> visited;
//queue<int> Q;
      Mesh * m;
      map < int, Simplex * >::iterator mIt;
   };

   class siterator {
public:
      siterator();
      siterator(Simplex * s);
      void operator++();
      void operator++(int);
      siterator& operator=(const siterator & it);
      bool operator==(const siterator & it);
      bool operator!=(const siterator & it);
      Simplex * operator*();

private:
      set < Simplex * >V;
      deque < Simplex * >Q;
   };

   siterator sbegin();
   siterator send();


   class piterator {
public:
      piterator();
      piterator(map< int, Simplex *>::iterator m);
      void operator++();
      void operator++(int);
      piterator& operator=(const piterator & it);
      bool operator==(const piterator & it);
      bool operator!=(const piterator & it);
      int operator*();

private:
      map < int, Simplex * >::iterator mit;

   };

   piterator pbegin();
   piterator pend();

private:


};

#endif              // !defined(__MESH__)
