#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#define INEXACT                 /* Nothing */

#define REAL double             /* float or double */
#define REALPRINT doubleprint
#define REALRAND doublerand
#define NARROWRAND narrowdoublerand
#define UNIFORMRAND uniformdoublerand

/* Which of the following two methods of finding the absolute values is      */
/*   fastest is compiler-dependent.  A few compilers can inline and optimize */
/*   the fabs() call; but most will incur the overhead of a function call,   */
/*   which is disastrously slow.  A faster way on IEEE machines might be to  */
/*   mask the appropriate bit, but that's difficult to do in C.              */

#define Absolute(a)  ((a) >= 0.0 ? (a) : -(a))
/* #define Absolute(a)  fabs(a) */

/* Many of the operations are broken up into two pieces, a main part that    */
/*   performs an approximate operation, and a "tail" that computes the       */
/*   roundoff error of that operation.                                       */
/*                                                                           */
/* The operations Fast_Two_Sum(), Fast_Two_Diff(), Two_Sum(), Two_Diff(),    */
/*   Split(), and Two_Product() are all implemented as described in the      */
/*   reference.  Each of these macros requires certain variables to be       */
/*   defined in the calling routine.  The variables `bvirt', `c', `abig',    */
/*   `_i', `_j', `_k', `_l', `_m', and `_n' are declared `INEXACT' because   */
/*   they store the result of an operation that may incur roundoff error.    */
/*   The input parameter `x' (or the highest numbered `x_' parameter) must   */
/*   also be declared `INEXACT'.                                             */

#define Fast_Two_Sum_Tail(a, b, x, y) \
   bvirt = x - a; \
   y = b - bvirt

#define Fast_Two_Sum(a, b, x, y) \
   x = (REAL) (a + b); \
   Fast_Two_Sum_Tail(a, b, x, y)

#define Fast_Two_Diff_Tail(a, b, x, y) \
   bvirt = a - x; \
   y = bvirt - b

#define Fast_Two_Diff(a, b, x, y) \
   x = (REAL) (a - b); \
   Fast_Two_Diff_Tail(a, b, x, y)

#define Two_Sum_Tail(a, b, x, y) \
   bvirt = (REAL) (x - a); \
   avirt = x - bvirt; \
   bround = b - bvirt; \
   around = a - avirt; \
   y = around + bround

#define Two_Sum(a, b, x, y) \
   x = (REAL) (a + b); \
   Two_Sum_Tail(a, b, x, y)

#define Two_Diff_Tail(a, b, x, y) \
   bvirt = (REAL) (a - x); \
   avirt = x + bvirt; \
   bround = bvirt - b; \
   around = a - avirt; \
   y = around + bround

#define Two_Diff(a, b, x, y) \
   x = (REAL) (a - b); \
   Two_Diff_Tail(a, b, x, y)

#define Split(a, ahi, alo) \
   c = (REAL) (splitter * a); \
   abig = (REAL) (c - a); \
   ahi = c - abig; \
   alo = a - ahi

#define Two_Product_Tail(a, b, x, y) \
   Split(a, ahi, alo); \
   Split(b, bhi, blo); \
   err1 = x - (ahi * bhi); \
   err2 = err1 - (alo * bhi); \
   err3 = err2 - (ahi * blo); \
   y = (alo * blo) - err3

#define Two_Product(a, b, x, y) \
   x = (REAL) (a * b); \
   Two_Product_Tail(a, b, x, y)

/* Two_Product_Presplit() is Two_Product() where one of the inputs has       */
/*   already been split.  Avoids redundant splitting.                        */

#define Two_Product_Presplit(a, b, bhi, blo, x, y) \
   x = (REAL) (a * b); \
   Split(a, ahi, alo); \
   err1 = x - (ahi * bhi); \
   err2 = err1 - (alo * bhi); \
   err3 = err2 - (ahi * blo); \
   y = (alo * blo) - err3

/* Two_Product_2Presplit() is Two_Product() where both of the inputs have    */
/*   already been split.  Avoids redundant splitting.                        */

#define Two_Product_2Presplit(a, ahi, alo, b, bhi, blo, x, y) \
   x = (REAL) (a * b); \
   err1 = x - (ahi * bhi); \
   err2 = err1 - (alo * bhi); \
   err3 = err2 - (ahi * blo); \
   y = (alo * blo) - err3

/* Square() can be done more quickly than Two_Product().                     */

#define Square_Tail(a, x, y) \
   Split(a, ahi, alo); \
   err1 = x - (ahi * ahi); \
   err3 = err1 - ((ahi + ahi) * alo); \
   y = (alo * alo) - err3

#define Square(a, x, y) \
   x = (REAL) (a * a); \
   Square_Tail(a, x, y)

/* Macros for summing expansions of various fixed lengths.  These are all    */
/*   unrolled versions of Expansion_Sum().                                   */

#define Two_One_Sum(a1, a0, b, x2, x1, x0) \
   Two_Sum(a0, b, _i, x0); \
   Two_Sum(a1, _i, x2, x1)

#define Two_One_Diff(a1, a0, b, x2, x1, x0) \
   Two_Diff(a0, b, _i, x0); \
   Two_Sum( a1, _i, x2, x1)

#define Two_Two_Sum(a1, a0, b1, b0, x3, x2, x1, x0) \
   Two_One_Sum(a1, a0, b0, _j, _0, x0); \
   Two_One_Sum(_j, _0, b1, x3, x2, x1)

#define Two_Two_Diff(a1, a0, b1, b0, x3, x2, x1, x0) \
   Two_One_Diff(a1, a0, b0, _j, _0, x0); \
   Two_One_Diff(_j, _0, b1, x3, x2, x1)

#define Four_One_Sum(a3, a2, a1, a0, b, x4, x3, x2, x1, x0) \
   Two_One_Sum(a1, a0, b, _j, x1, x0); \
   Two_One_Sum(a3, a2, _j, x4, x3, x2)

#define Four_Two_Sum(a3, a2, a1, a0, b1, b0, x5, x4, x3, x2, x1, x0) \
   Four_One_Sum(a3, a2, a1, a0, b0, _k, _2, _1, _0, x0); \
   Four_One_Sum(_k, _2, _1, _0, b1, x5, x4, x3, x2, x1)

#define Four_Four_Sum(a3, a2, a1, a0, b4, b3, b1, b0, x7, x6, x5, x4, x3, x2, \
                      x1, x0) \
   Four_Two_Sum(a3, a2, a1, a0, b1, b0, _l, _2, _1, _0, x1, x0); \
   Four_Two_Sum(_l, _2, _1, _0, b4, b3, x7, x6, x5, x4, x3, x2)

#define Eight_One_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b, x8, x7, x6, x5, x4, \
                      x3, x2, x1, x0) \
   Four_One_Sum(a3, a2, a1, a0, b, _j, x3, x2, x1, x0); \
   Four_One_Sum(a7, a6, a5, a4, _j, x8, x7, x6, x5, x4)

#define Eight_Two_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b1, b0, x9, x8, x7, \
                      x6, x5, x4, x3, x2, x1, x0) \
   Eight_One_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b0, _k, _6, _5, _4, _3, _2, \
                 _1, _0, x0); \
   Eight_One_Sum(_k, _6, _5, _4, _3, _2, _1, _0, b1, x9, x8, x7, x6, x5, x4, \
                 x3, x2, x1)

#define Eight_Four_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b4, b3, b1, b0, x11, \
                       x10, x9, x8, x7, x6, x5, x4, x3, x2, x1, x0) \
   Eight_Two_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b1, b0, _l, _6, _5, _4, _3, \
                 _2, _1, _0, x1, x0); \
   Eight_Two_Sum(_l, _6, _5, _4, _3, _2, _1, _0, b4, b3, x11, x10, x9, x8, \
                 x7, x6, x5, x4, x3, x2)

/* Macros for multiplying expansions of various fixed lengths.               */

#define Two_One_Product(a1, a0, b, x3, x2, x1, x0) \
   Split(b, bhi, blo); \
   Two_Product_Presplit(a0, b, bhi, blo, _i, x0); \
   Two_Product_Presplit(a1, b, bhi, blo, _j, _0); \
   Two_Sum(_i, _0, _k, x1); \
   Fast_Two_Sum(_j, _k, x3, x2)

#define Four_One_Product(a3, a2, a1, a0, b, x7, x6, x5, x4, x3, x2, x1, x0) \
   Split(b, bhi, blo); \
   Two_Product_Presplit(a0, b, bhi, blo, _i, x0); \
   Two_Product_Presplit(a1, b, bhi, blo, _j, _0); \
   Two_Sum(_i, _0, _k, x1); \
   Fast_Two_Sum(_j, _k, _i, x2); \
   Two_Product_Presplit(a2, b, bhi, blo, _j, _0); \
   Two_Sum(_i, _0, _k, x3); \
   Fast_Two_Sum(_j, _k, _i, x4); \
   Two_Product_Presplit(a3, b, bhi, blo, _j, _0); \
   Two_Sum(_i, _0, _k, x5); \
   Fast_Two_Sum(_j, _k, x7, x6)

#define Two_Two_Product(a1, a0, b1, b0, x7, x6, x5, x4, x3, x2, x1, x0) \
   Split(a0, a0hi, a0lo); \
   Split(b0, bhi, blo); \
   Two_Product_2Presplit(a0, a0hi, a0lo, b0, bhi, blo, _i, x0); \
   Split(a1, a1hi, a1lo); \
   Two_Product_2Presplit(a1, a1hi, a1lo, b0, bhi, blo, _j, _0); \
   Two_Sum(_i, _0, _k, _1); \
   Fast_Two_Sum(_j, _k, _l, _2); \
   Split(b1, bhi, blo); \
   Two_Product_2Presplit(a0, a0hi, a0lo, b1, bhi, blo, _i, _0); \
   Two_Sum(_1, _0, _k, x1); \
   Two_Sum(_2, _k, _j, _1); \
   Two_Sum(_l, _j, _m, _2); \
   Two_Product_2Presplit(a1, a1hi, a1lo, b1, bhi, blo, _j, _0); \
   Two_Sum(_i, _0, _n, _0); \
   Two_Sum(_1, _0, _i, x2); \
   Two_Sum(_2, _i, _k, _1); \
   Two_Sum(_m, _k, _l, _2); \
   Two_Sum(_j, _n, _k, _0); \
   Two_Sum(_1, _0, _j, x3); \
   Two_Sum(_2, _j, _i, _1); \
   Two_Sum(_l, _i, _m, _2); \
   Two_Sum(_1, _k, _i, x4); \
   Two_Sum(_2, _i, _k, x5); \
   Two_Sum(_m, _k, x7, x6)

/* An expansion of length two can be squared more quickly than finding the   */
/*   product of two different expansions of length two, and the result is    */
/*   guaranteed to have no more than six (rather than eight) components.     */

#define Two_Square(a1, a0, x5, x4, x3, x2, x1, x0) \
   Square(a0, _j, x0); \
   _0 = a0 + a0; \
   Two_Product(a1, _0, _k, _1); \
   Two_One_Sum(_k, _1, _j, _l, _2, x1); \
   Square(a1, _j, _1); \
   Two_Two_Sum(_j, _1, _l, _2, x5, x4, x3, x2)



#define One_Two_Diff(b, a1, a0, x2, x1, x0) \
   Two_Diff(b, a0, _i, x0); \
   Two_Diff(_i, a1, x2, x1);

#define Two_Absolute(a1, a0, x1, x0) \
   if (a1 >= 0) {x1 = a1; x0 = a0}; \
                 else {x1 = -a1; x0 = -a1}

#define Two_UMinus(a1, a0, x1, x0) \
   x1 = -a1; x0 = -a0


REAL splitter;                  /* = 2^ceiling(p / 2) + 1.  Used to split floats in half. */
REAL epsilon;                   /* = 2^(-p).  Used to estimate roundoff errors. */
/* A set of coefficients used to calculate maximum roundoff errors.          */
REAL resulterrbound;
REAL ccwerrboundA, ccwerrboundB, ccwerrboundC;
REAL o3derrboundA, o3derrboundB, o3derrboundC;
REAL iccerrboundA, iccerrboundB, iccerrboundC;
REAL isperrboundA, isperrboundB, isperrboundC;

void doubleprint(double number);
void floatprint(float number);
void expansion_print(int elen, REAL * e);
double doublerand();
double narrowdoublerand();
double uniformdoublerand();
float floatrand();
float narrowfloatrand();
float uniformfloatrand();
void exactinit();
int grow_expansion(int elen, REAL * e, REAL b, REAL * h);
int grow_expansion_zeroelim(int elen, REAL * e, REAL b, REAL * h);
int expansion_sum(int elen, REAL * e, int flen, REAL * f, REAL * h);
int expansion_sum_zeroelim1(int elen, REAL * e, int flen, REAL * f,
                            REAL * h);
int expansion_sum_zeroelim2(int elen, REAL * e, int flen, REAL * f,
                            REAL * h);
int fast_expansion_sum(int elen, REAL * e, int flen, REAL * f, REAL * h);
int fast_expansion_sum_zeroelim(int elen, REAL * e, int flen, REAL * f,
                                REAL * h);
int linear_expansion_sum(int elen, REAL * e, int flen, REAL * f, REAL * h);
int linear_expansion_sum_zeroelim(int elen, REAL * e, int flen, REAL * f,
                                  REAL * h);
int scale_expansion(int elen, REAL * e, REAL b, REAL * h);
int scale_expansion_zeroelim(int elen, REAL * e, REAL b, REAL * h);
int compress(int elen, REAL * e, REAL * h);
REAL estimate(int elen, REAL * e);
int shrink_expansion(REAL b, int elen, REAL * e, REAL * h);
int fast_expansion_diff_zeroelim(int elen, REAL * e, int flen, REAL * f,
                                 REAL * h);
int mult_expansion_zeroelim(int elen, REAL * e, int blen, REAL * b,
                            REAL * h);
int expansion_double(int elen, REAL * e, REAL * h);
int expansion_absolute(int elen, REAL * e, REAL * h);
int expansion_uminus(int elen, REAL * e, REAL * h);
