/*****************************************************************************/
/* Additional Routines for Arbitrary Precision Floating-point Arithmetic     */
/*                                                                           */
/* Needed for the Expression Compiler                                        */
/*                                                                           */
/* Aleksandar Nanevski                                                       */
/*****************************************************************************/

#include "predicates.h"

int shrink_expansion(REAL b, int elen, REAL * e, REAL * h)
{                               /* e and h can be the same. */
   REAL Q;
   INEXACT REAL Qnew;
   int eindex;
   REAL enow;
   INEXACT REAL bvirt;
   REAL avirt, bround, around;

   Q = b;
   for (eindex = 0; eindex < elen; eindex++) {
      enow = e[eindex];
      Two_Diff(Q, enow, Qnew, h[eindex]);
      Q = Qnew;
   }
   h[eindex] = Q;
   return eindex + 1;
}



int fast_expansion_diff_zeroelim(int elen, REAL * e, int flen, REAL * f,
                                 REAL * h)
{                               /* h cannot be e or f */
   REAL Q;
   INEXACT REAL Qnew;
   INEXACT REAL hh;
   INEXACT REAL bvirt;
   REAL avirt, bround, around;
   int eindex, findex, hindex;
   REAL enow, fnow;

   enow = e[0];
   fnow = -f[0];
   eindex = findex = 0;
   if ((fnow > enow) == (fnow > -enow)) {
      Q = enow;
      enow = e[++eindex];
   } else {
      Q = fnow;
      fnow = -f[++findex];
   }
   hindex = 0;
   if ((eindex < elen) && (findex < flen)) {
      if ((fnow > enow) == (fnow > -enow)) {
         Fast_Two_Sum(enow, Q, Qnew, hh);
         enow = e[++eindex];
      } else {
         Fast_Two_Sum(fnow, Q, Qnew, hh);
         fnow = -f[++findex];
      }
      Q = Qnew;
      if (hh != 0.0) {
         h[hindex++] = hh;
      }
      while ((eindex < elen) && (findex < flen)) {
         if ((fnow > enow) == (fnow > -enow)) {
            Two_Sum(Q, enow, Qnew, hh);
            enow = e[++eindex];
         } else {
            Two_Sum(Q, fnow, Qnew, hh);
            fnow = -f[++findex];
         }
         Q = Qnew;
         if (hh != 0.0) {
            h[hindex++] = hh;
         }
      }
   }
   while (eindex < elen) {
      Two_Sum(Q, enow, Qnew, hh);
      enow = e[++eindex];
      Q = Qnew;
      if (hh != 0.0) {
         h[hindex++] = hh;
      }
   }
   while (findex < flen) {
      Two_Sum(Q, fnow, Qnew, hh);
      fnow = -f[++findex];
      Q = Qnew;
      if (hh != 0.0) {
         h[hindex++] = hh;
      }
   }
   if ((Q != 0.0) || (hindex == 0)) {
      h[hindex++] = Q;
   }
   return hindex;
}



int mult_expansion_zeroelim(int elen, REAL * e, int blen, REAL * b,
                            REAL * h)
{                               /* h cannot be e nor b */
   int i, j, k, tlen, slen, llen, newt_len;
   int *l;
   REAL **t, *newt;

   /* array of pointers to blen expansions */
   tlen = blen * sizeof(REAL *);
   t = (REAL **) malloc(tlen);

   /* array of lengths of blen expansions */
   llen = blen * sizeof(int);
   l = (int *) malloc((size_t) llen);

   /* size of the scale_expansion(e, b[i]) */
   slen = 2 * elen * sizeof(REAL);


   /* initialize */
   for (i = 0; i < blen; i++) {
      t[i] = (REAL *) malloc((size_t) slen);
      l[i] = scale_expansion_zeroelim(elen, e, b[i], t[i]);
   }

   /* distill */
   for (i = blen; i > 1; i = (i >> 1) + (i & 1)) {
      for (j = k = 0; j + 1 < i; j += 2, k++) {
         newt =
            (REAL *) malloc((size_t) (l[j] + l[j + 1]) *
                            sizeof(double));
         newt_len =
            fast_expansion_sum_zeroelim(l[j], t[j], l[j + 1], t[j + 1],
                                        newt);
         free(t[j]);
         free(t[j + 1]);
         t[k] = newt;
         l[k] = newt_len;
      }

      if (j + 1 == i) {
         t[k] = t[j];
         l[k] = l[j];
      }
   }

   /* copy the result */
   for (i = 0; i < l[0]; i++)
      h[i] = t[0][i];
   newt_len = l[0];

   free(t[0]);
   free(l);
   free(t);

   return newt_len;
}

/*


   REAL s[256], t[256];

   h[0] = 0;

   for (eindex = 0; eindex < elen; eindex++) {
    enow = e[eindex];
    slen = scale_expansion(blen, b, enow, s);
    tlen = fast_expansion_sum(hlen, h, slen, s, t);
    h[0] = hlen = 0;
    for (tindex = 0; tindex < tlen; tindex++) {
      tt = t[tindex];
      if (tt != 0) {
        h[hlen++] = tt;
      }
    }
   }

   return hlen;
   }

 */



int expansion_double(int elen, REAL * e, REAL * h)
{
   int eindex;
   h[0] = 0;
   for (eindex = 0; eindex < elen; eindex++) {
      h[eindex] = 2.0 * e[eindex];
   }
   return elen;
}



int expansion_absolute(int elen, REAL * e, REAL * h)
{
   int eindex;
   if (e[0] >= 0)
      for (eindex = 0; eindex < elen; eindex++)
         h[eindex] = e[eindex];
   else
      for (eindex = 0; eindex < elen; eindex++)
         h[eindex] = -e[eindex];

   return elen;
}



int expansion_uminus(int elen, REAL * e, REAL * h)
{
   int eindex;
   for (eindex = 0; eindex < elen; eindex++)
      h[eindex] = -e[eindex];
   return elen;
}



/*****************************************************************************/
/* From this point on, functions are copied from Jonathan Shewchuk's         */
/* robust predicates.  Only the four functions below are needed.             */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/*  fast_expansion_sum_zeroelim()   Sum two expansions, eliminating zero     */
/*                                  components from the output expansion.    */
/*                                                                           */
/*  Sets h = e + f.  See the long version of my paper for details.           */
/*                                                                           */
/*  If round-to-even is used (as with IEEE 754), maintains the strongly      */
/*  nonoverlapping property.  (That is, if e is strongly nonoverlapping, h   */
/*  will be also.)  Does NOT maintain the nonoverlapping or nonadjacent      */
/*  properties.                                                              */
/*                                                                           */
/*****************************************************************************/

int fast_expansion_sum_zeroelim(int elen, REAL * e, int flen, REAL * f,
                                REAL * h)
{                               /* h cannot be e or f. */
   REAL Q;
   INEXACT REAL Qnew;
   INEXACT REAL hh;
   INEXACT REAL bvirt;
   REAL avirt, bround, around;
   int eindex, findex, hindex;
   REAL enow, fnow;

   enow = e[0];
   fnow = f[0];
   eindex = findex = 0;
   if ((fnow > enow) == (fnow > -enow)) {
      Q = enow;
      enow = e[++eindex];
   } else {
      Q = fnow;
      fnow = f[++findex];
   }
   hindex = 0;
   if ((eindex < elen) && (findex < flen)) {
      if ((fnow > enow) == (fnow > -enow)) {
         Fast_Two_Sum(enow, Q, Qnew, hh);
         enow = e[++eindex];
      } else {
         Fast_Two_Sum(fnow, Q, Qnew, hh);
         fnow = f[++findex];
      }
      Q = Qnew;
      if (hh != 0.0) {
         h[hindex++] = hh;
      }
      while ((eindex < elen) && (findex < flen)) {
         if ((fnow > enow) == (fnow > -enow)) {
            Two_Sum(Q, enow, Qnew, hh);
            enow = e[++eindex];
         } else {
            Two_Sum(Q, fnow, Qnew, hh);
            fnow = f[++findex];
         }
         Q = Qnew;
         if (hh != 0.0) {
            h[hindex++] = hh;
         }
      }
   }
   while (eindex < elen) {
      Two_Sum(Q, enow, Qnew, hh);
      enow = e[++eindex];
      Q = Qnew;
      if (hh != 0.0) {
         h[hindex++] = hh;
      }
   }
   while (findex < flen) {
      Two_Sum(Q, fnow, Qnew, hh);
      fnow = f[++findex];
      Q = Qnew;
      if (hh != 0.0) {
         h[hindex++] = hh;
      }
   }
   if ((Q != 0.0) || (hindex == 0)) {
      h[hindex++] = Q;
   }
   return hindex;
}


/*****************************************************************************/
/*                                                                           */
/*  exactinit()   Initialize the variables used for exact arithmetic.        */
/*                                                                           */
/*  `epsilon' is the largest power of two such that 1.0 + epsilon = 1.0 in   */
/*  floating-point arithmetic.  `epsilon' bounds the relative roundoff       */
/*  error.  It is used for floating-point error analysis.                    */
/*                                                                           */
/*  `splitter' is used to split floating-point numbers into two half-        */
/*  length significands for exact multiplication.                            */
/*                                                                           */
/*  I imagine that a highly optimizing compiler might be too smart for its   */
/*  own good, and somehow cause this routine to fail, if it pretends that    */
/*  floating-point arithmetic is too much like real arithmetic.              */
/*                                                                           */
/*  Don't change this routine unless you fully understand it.                */
/*                                                                           */
/*****************************************************************************/

void exactinit()
{
   REAL half;
   REAL check, lastcheck;
   int every_other;

   every_other = 1;
   half = 0.5;
   epsilon = 1.0;
   splitter = 1.0;
   check = 1.0;
   /* Repeatedly divide `epsilon' by two until it is too small to add to    */
   /*   one without causing roundoff.  (Also check if the sum is equal to   */
   /*   the previous sum, for machines that round up instead of using exact */
   /*   rounding.  Not that this library will work on such machines anyway. */
   do {
      lastcheck = check;
      epsilon *= half;
      if (every_other) {
         splitter *= 2.0;
      }
      every_other = !every_other;
      check = 1.0 + epsilon;
   } while ((check != 1.0) && (check != lastcheck));
   splitter += 1.0;

   /* Error bounds for orientation and incircle tests. */
   resulterrbound = (3.0 + 8.0 * epsilon) * epsilon;
   ccwerrboundA = (3.0 + 16.0 * epsilon) * epsilon;
   ccwerrboundB = (2.0 + 12.0 * epsilon) * epsilon;
   ccwerrboundC = (9.0 + 64.0 * epsilon) * epsilon * epsilon;
   o3derrboundA = (7.0 + 56.0 * epsilon) * epsilon;
   o3derrboundB = (3.0 + 28.0 * epsilon) * epsilon;
   o3derrboundC = (26.0 + 288.0 * epsilon) * epsilon * epsilon;
   iccerrboundA = (10.0 + 96.0 * epsilon) * epsilon;
   iccerrboundB = (4.0 + 48.0 * epsilon) * epsilon;
   iccerrboundC = (44.0 + 576.0 * epsilon) * epsilon * epsilon;
   isperrboundA = (16.0 + 224.0 * epsilon) * epsilon;
   isperrboundB = (5.0 + 72.0 * epsilon) * epsilon;
   isperrboundC = (71.0 + 1408.0 * epsilon) * epsilon * epsilon;
}

/*****************************************************************************/
/*                                                                           */
/*  scale_expansion_zeroelim()   Multiply an expansion by a scalar,          */
/*                               eliminating zero components from the        */
/*                               output expansion.                           */
/*                                                                           */
/*  Sets h = be.  See either version of my paper for details.                */
/*                                                                           */
/*  Maintains the nonoverlapping property.  If round-to-even is used (as     */
/*  with IEEE 754), maintains the strongly nonoverlapping and nonadjacent    */
/*  properties as well.  (That is, if e has one of these properties, so      */
/*  will h.)                                                                 */
/*                                                                           */
/*****************************************************************************/

int scale_expansion_zeroelim(int elen, REAL * e, REAL b, REAL * h)
{                               /* e and h cannot be the same. */
   INEXACT REAL Q, sum;
   REAL hh;
   INEXACT REAL product1;
   REAL product0;
   int eindex, hindex;
   REAL enow;
   INEXACT REAL bvirt;
   REAL avirt, bround, around;
   INEXACT REAL c;
   INEXACT REAL abig;
   REAL ahi, alo, bhi, blo;
   REAL err1, err2, err3;

   Split(b, bhi, blo);
   Two_Product_Presplit(e[0], b, bhi, blo, Q, hh);
   hindex = 0;
   if (hh != 0) {
      h[hindex++] = hh;
   }
   for (eindex = 1; eindex < elen; eindex++) {
      enow = e[eindex];
      Two_Product_Presplit(enow, b, bhi, blo, product1, product0);
      Two_Sum(Q, product0, sum, hh);
      if (hh != 0) {
         h[hindex++] = hh;
      }
      Fast_Two_Sum(product1, sum, Q, hh);
      if (hh != 0) {
         h[hindex++] = hh;
      }
   }
   if ((Q != 0.0) || (hindex == 0)) {
      h[hindex++] = Q;
   }
   return hindex;
}


/*****************************************************************************/
/*                                                                           */
/*  estimate()   Produce a one-word estimate of an expansion's value.        */
/*                                                                           */
/*  See either version of my paper for details.                              */
/*                                                                           */
/*****************************************************************************/

REAL estimate(int elen, REAL * e)
{
   REAL Q;
   int eindex;

   Q = e[0];
   for (eindex = 1; eindex < elen; eindex++) {
      Q += e[eindex];
   }
   return Q;
}
